#include <abort.h>
#include <sys/region.h>
#include "tree.h"

inline void init_TreeNode( TreeNode* self, void* data ) {
	init_Vec( &self->children, sizeof(TreeNode*), ALIGN_DEFAULT );
	self->parent = NULL;
	self->data = data;
}

inline void destroy_TreeNode( TreeNode* self ) {
	destroy_Vec( &self->children );
}

inline void attachChild_TreeNode( TreeNode* self, TreeNode* child ) {
	pushBack_Vec( &self->children, &child );
	child->parent = self;
}

inline uint_t getNumChildren_TreeNode( const TreeNode* self ) {
	return getLength_Vec( &self->children );
}

inline TreeNode* getParent_TreeNode( const TreeNode* self ) {
	return self->parent;
}

inline TreeNode* getChild_TreeNode( const TreeNode* self, uint_t i ) {
	return *(TreeNode* const *)getConstElement_Vec( &self->children, i );
}

inline void* getData_TreeNode( const TreeNode* self ) {
	return self->data;
}

inline void init_TreeNodeIterator( TreeNodeIterator* self ) {
	init_Stack( &self->stack, sizeof(self->i) );
	reset_TreeNodeIterator( self, NULL );
}

inline void destroy_TreeNodeIterator( TreeNodeIterator* self ) {
	destroy_Stack( &self->stack );
}

TreeNode* inc_TreeNodeIterator( TreeNodeIterator* self ) {
	TreeNode* node;
i_believe_in_goto:	node = self->next;
	if( NULL == node ) {
		return NULL;
	}

	bool node_visited = self->i;

	uint_t n = getNumChildren_TreeNode( node );
	if( self->i < n ) {
		self->next = getChild_TreeNode( node, self->i );
		self->i++;
		push_Stack( &self->stack, &self->i );

		self->i = 0;
	} else {
		if( true == isEmpty_Stack( &self->stack ) ) {
			if( 0 != n ) {
				return NULL;
			}

			/* it was a single-node tree */
			self->next = NULL;
			return self->node;
		}
	
		self->i = *(uint_t*)pop_Stack( &self->stack );
		self->next = getParent_TreeNode( node );
	}

	if( node_visited ) {
		goto i_believe_in_goto;
	}

	return node;
}

inline void reset_TreeNodeIterator( TreeNodeIterator* self, TreeNode* node ) {
	self->node = node;
	self->next = self->node;
	self->i = 0;
}

