#include <ultra64.h>
#include "matrix.h"

inline void identity_Matrix( Matrix* mtx ) {
	guMtxIdentF( mtx->m );
}

inline void rotate_Matrix( Matrix* mtx, float a, float x, float y, float z ) {
	guRotateF( mtx->m, a, x, y, z );
}

inline void scale_Matrix( Matrix* mtx, float x, float y, float z ) {
	guScaleF( mtx->m, x, y, z );
}

inline void translate_Matrix( Matrix* mtx, float x, float y, float z ) {
	guTranslateF( mtx->m, x, y, z );
}

void invert_Matrix( const Matrix* restrict mi, Matrix* restrict mtxout ) {
	float det = MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 3)
		  + MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 1)
		  + MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 2)
		  + MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 2)
		  
		  + MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 3)
		  + MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 0)
		  + MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 3)
		  + MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 0)
		  
		  + MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 1)
		  + MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 1)
		  + MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 2)
		  + MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 0)
		  
		  - MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 2)
		  - MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 3)
		  - MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 1)
		  - MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 3)
		  
		  - MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 0)
		  - MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 2)
		  - MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 1)
		  - MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 3)
		  
		  - MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 0)
		  - MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 2)
		  - MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 0)
		  - MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 1);
	
	MATRIX_GET(*mtxout, 0, 0) = ( MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 3)
			+ MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 1)
			+ MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 2)
			- MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi ,3, 2)
			- MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 3)
			- MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 1) ) / det;

	MATRIX_GET(*mtxout, 0, 1) = ( MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 2)
			+ MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 3)
			+ MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 1)
			- MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 3)
			- MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 1)
			- MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 2) ) / det;

	MATRIX_GET(*mtxout, 0, 2) = ( MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 3, 3)
			+ MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 3, 1)
			+ MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 3, 2)
			- MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 3, 2)
			- MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 3, 3)
			- MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 3, 1) ) / det;

	MATRIX_GET(*mtxout, 0, 3) = ( MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 2)
			+ MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 3)
			+ MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 1)
			- MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 3)
			- MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 1)
			- MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 2) ) / det;

	MATRIX_GET(*mtxout, 1, 0) = ( MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 2)
			+ MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 3)
			+ MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 0)
			- MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 3)
			- MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 0)
			- MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 2) ) / det;

	MATRIX_GET(*mtxout, 1, 1) = ( MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 3)
			+ MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 0)
			+ MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 2)
			- MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 2)
			- MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 3)
			- MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 0) ) / det;

	MATRIX_GET(*mtxout, 1, 2) = ( MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 3, 2)
			+ MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 3, 3)
			+ MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 3, 0)
			- MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 3, 3)
			- MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 3, 0)
			- MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 3, 2) ) / det;

	MATRIX_GET(*mtxout, 1, 3) = ( MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 3)
			+ MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 0)
			+ MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 2)
			- MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 2)
			- MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 3)
			- MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 0) ) / det;

	MATRIX_GET(*mtxout, 2, 0) = ( MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 3)
			+ MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 0)
			+ MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 1)
			- MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 1)
			- MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 3)
			- MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 0) ) / det;

	MATRIX_GET(*mtxout, 2, 1) = ( MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 1)
			+ MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 3)
			+ MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 0)
			- MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 3)
			- MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 2, 3)*MATRIX_GET(*mi, 3, 0)
			- MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 1) ) / det;

	MATRIX_GET(*mtxout, 2, 2) = ( MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 3, 3)
			+ MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 3, 0)
			+ MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 3, 1)
			- MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 3, 1)
			- MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 3, 3)
			- MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 3, 0) ) / det;

	MATRIX_GET(*mtxout, 2, 3) = ( MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 1)
			+ MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 3)
			+ MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 0)
			- MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 3)
			- MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 3)*MATRIX_GET(*mi, 2, 0)
			- MATRIX_GET(*mi, 0, 3)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 1) ) / det;

	MATRIX_GET(*mtxout, 3, 0) = ( MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 1)
			+ MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 2)
			+ MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 0)
			- MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 2)
			- MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 0)
			- MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 1) ) / det;

	MATRIX_GET(*mtxout, 3, 1) = ( MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 2)
			+ MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 0)
			+ MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 1)
			- MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 2, 2)*MATRIX_GET(*mi, 3, 1)
			- MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 2, 0)*MATRIX_GET(*mi, 3, 2)
			- MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 2, 1)*MATRIX_GET(*mi, 3, 0) ) / det;

	MATRIX_GET(*mtxout, 3, 2) = ( MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 3, 1)
			+ MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 3, 2)
			+ MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 3, 0)
			- MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 3, 2)
			- MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 3, 0)
			- MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 3, 1) ) / det;

	MATRIX_GET(*mtxout, 3, 3) = ( MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 2)
			+ MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 0)
			+ MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 1)
			- MATRIX_GET(*mi, 0, 0)*MATRIX_GET(*mi, 1, 2)*MATRIX_GET(*mi, 2, 1)
			- MATRIX_GET(*mi, 0, 1)*MATRIX_GET(*mi, 1, 0)*MATRIX_GET(*mi, 2, 2)
			- MATRIX_GET(*mi, 0, 2)*MATRIX_GET(*mi, 1, 1)*MATRIX_GET(*mi, 2, 0) ) / det;
}

inline void multiply_Matrix( Matrix* out, Matrix* ma, Matrix* mb ) {
	guMtxCatF( ma->m, mb->m, out->m );
}

inline void perspective_Matrix( Matrix* mtx, float* perspNorm, float fovy, float aspect, float near, float far ) {
	u16 u16_norm;
	guPerspectiveF(
		mtx->m,
		&u16_norm,
		fovy,
		aspect,
		near,
		far,
		1.0f
	);
	*perspNorm = (float)u16_norm;
}

inline void align_Matrix( Matrix* mtx, float angle, float x, float y, float z ) {
	guAlignF( mtx->m, angle, x, y, z );
}

