#include <abort.h>
#include <malloc.h>
#include <ultra64.h>
#include "dlist.h"

void init_DList( DList* self ) {
	self->size = 8192;
	self->data = malloc( self->size );
	reset_DList( self );
}

void destroy_DList( DList* self ) {
	free( self->data );
}

inline void reset_DList( DList* self ) {
	self->current = self->data;
}

inline void* getData_DList( const DList* self ) {
	return self->data;
}

inline size_t getSize_DList( const DList* self ) {
	return self->size;
}

#define FOO(dl, T)			\
	T* p = (T*)((dl)->current);	\
	++(*(T**)(&(dl)->current));	\
	if((uint8_t*)dl->current >= (uint8_t*)dl->data + dl->size)	abort()

inline void pushStart_DList( DList* dlist ) {
	FOO(dlist, Command_Start);
	p->name = Start;
}

inline void pushSetVertexSource_DList( DList* dlist, const void* vts, uint8_t format ) {
	FOO(dlist, Command_SetVertexSource);
	p->name = SetVertexSource;
	p->format = format;
	p->data = vts;
}

inline void pushSetIndexSource_DList( DList* dlist, const void* ids ) {
	FOO(dlist, Command_SetIndexSource);
	p->name = SetIndexSource;
	p->data = ids;
}

inline void pushLoadVertexBuffer_DList( DList* dlist, uint32_t src_ofs,
	uint32_t dst_ofs, uint32_t nvts ) {

	FOO(dlist, Command_LoadVertexBuffer);
	p->name = LoadVertexBuffer;
	p->src_ofs = src_ofs;
	p->dst_ofs = dst_ofs;
	p->nvts = nvts;
}

inline void pushLoadIndexBuffer_DList( DList* dlist, uint32_t src_ofs,
	uint32_t dst_ofs, uint32_t nids ) {

	FOO(dlist, Command_LoadIndexBuffer);
	p->name = LoadIndexBuffer;
	p->src_ofs = src_ofs;
	p->dst_ofs = dst_ofs;
	p->nids = nids;
}

inline void pushDrawPrimitives_DList( DList* dlist, uint8_t type, uint32_t num ) {
	FOO(dlist, Command_DrawPrimitives);
	p->name = DrawPrimitives;
	p->type = type;
	p->num = num;
}

inline void pushSetBuffer_DList( DList* dlist, uint8_t buftype, uint8_t format,
	uint16_t width, uint16_t height, void* buf ) {

	FOO(dlist, Command_SetBuffer);
	p->name = SetBuffer;
	p->buftype = buftype;
	p->format = format;
	p->width = width;
	p->height = height;
	p->buf = buf;
}

inline void pushClearBuffers_DList( DList* dlist, uint8_t types ) {
	FOO(dlist, Command_ClearBuffers);
	p->name = ClearBuffers;
	p->types = types;
}

inline void pushSetClearColor_DList( DList* dlist, float r, float g, float b, float a ) {
	FOO(dlist, Command_SetClearColor);
	p->name = SetClearColor;
	p->r = r;
	p->g = g;
	p->b = b;
	p->a = a;
}

inline void pushSetClearZValue_DList( DList* dlist, float val ) {
	FOO(dlist, Command_SetClearZValue);
	p->name = SetClearZValue;
	p->val = val;
}

inline void pushLoadTexture_DList( DList* dlist, const Texture* texture ) {
	FOO(dlist, Command_LoadTexture);
	p->name = LoadTexture;
	p->texture = texture;
}

inline void pushJump_DList( DList* dlist, const DList* dl ) {
	FOO(dlist, Command_Jump);
	p->name = Jump;
	p->dl = dl;
}

inline void pushEnd_DList( DList* dlist ) {
	FOO(dlist, Command_End);
	p->name = End;
}

inline void pushSetTextureScale_DList( DList* dlist, float s, float t) {
	FOO(dlist, Command_SetTextureScale);
	p->name = SetTextureScale;
	p->s = s;
	p->t = t;
}

inline void pushLoadMatrix_DList( DList* dlist, Matrix* mtx, MatrixType type ) {
	FOO(dlist, Command_LoadMatrix);
	p->name = LoadMatrix;
	p->mtx = mtx;
	p->type = type;
}

inline void pushLoadLight_DList( DList* dlist, const Lux* light,
	const Vector3f* position, uint16_t idx ) {

	FOO(dlist, Command_LoadLight);
	p->name = LoadLight;
	p->light = light;
	p->pos = position;
	p->idx = idx;
}

inline void pushSync_DList( DList* dlist ) {
	FOO(dlist, Command_Sync);
	p->name = Sync;
}

inline void pushPipeSync_DList( DList* dlist ) {
	FOO(dlist, Command_PipeSync);
	p->name = PipeSync;
}

inline void pushJumpToTranslated_DList( DList* dlist, const void* dl ) {
	FOO(dlist, Command_JumpToTranslated);
	p->name = JumpToTranslated;
	p->dl = dl;
}

inline void pushSetCombineMode_DList( DList* dlist, CombineMode mode ) {
	FOO(dlist, Command_SetCombineMode);
	p->name = SetCombineMode;
	p->mode = mode;
}

inline void pushSetRenderMode_DList( DList* dlist, RenderMode mode ) {
	FOO(dlist, Command_SetRenderMode);
	p->name = SetRenderMode;
	p->mode = mode;
}

inline void pushNormalizePerspective_DList( DList* dlist, float scale ) {
	FOO(dlist, Command_NormalizePerspective);
	p->name = NormalizePerspective;
	p->scale = scale;
}

inline void pushSetVertexScale_DList( DList* dlist, float scale ) {
	FOO(dlist, Command_SetVertexScale);
	p->name = SetVertexScale;
	p->scale = scale;
}

inline void pushSetLookAt_DList( DList* dlist, LookAt* lookat ) {
	FOO(dlist, Command_SetLookAt);
	p->name = SetLookAt;
	p->lookat = lookat;
}

inline void pushSetGeometryMode_DList( DList* dlist, GeometryMode gmode ) {
	FOO(dlist, Command_SetGeometryMode);
	p->name = SetGeometryMode;
	p->gmode = gmode;
}

inline void pushClearGeometryMode_DList( DList* dlist, GeometryMode gmode ) {
	FOO(dlist, Command_ClearGeometryMode);
	p->name = ClearGeometryMode;
	p->gmode = gmode;
}

inline void pushSetViewport_DList( DList* dlist, const Viewport* vp ) {
	FOO(dlist, Command_SetViewport);
	p->name = SetViewport;
	p->vp = vp;
}

inline void pushSetMaterial_DList( DList* dlist, const Material* material ) {
	FOO(dlist, Command_SetMaterial);
	p->name = SetMaterial;
	p->mtl = material;
}

