#include <abort.h>
#include <actmgr/actmgr.h>
#include <containers/tree.h>
#include <engine.h>
#include <interpol/interpol.h>
#include <postproc/postproc.h>
#include <resmgr/resmgr.h>
#include <stdlib.h>
#include <sys/sys.h>

// BUG: display flickers if BILLBOARDS_NUM is too high (on PJ64); it doesn't if we don't iterate trough 2nd pass in engine.c
// Looks like the problem has something to do with length of the display list. On the other hand, CEN64 doesn't complain. Hm.
#define BILLBOARDS_NUM 10
#define SCREEN_ASPECT_RATIO	(320.0f/240.0f)
//#define POSTPROC_BLUR

static Engine	engine;
static float	angle_torus;	/* in degrees */

static struct {
	Vector3f	pos;
	Vector3f	dir;
	float		angle;	/* degrees */
} player = {
	.pos = VECTOR3F_SET( 0.0f, 60.0f, 0.0f ),
	.dir = VECTOR3F_SET( 0.0f, 0.0f, 0.0f ),
	.angle = 0.0f
};

void handle_vbi( void* arg ) {
	if( arg ) {}
	angle_torus += 1.0f;
}

static void handle_controller( void* arg ) {
	static float sty_old;
	static float stx_old;
	static float ud_old;
	if( arg ) {
		OSContPad* rdata = arg;
		player.angle -= 0.04f * stx_old + rdata->stick_x * 0.015f;

		player.dir.x = cosf( M_PI * (player.angle + 90) / 180 );
		//player.dir_y = player.dir.y;
		player.dir.z = -sinf( M_PI * (player.angle + 90) / 180 );

		player.pos.x += 0.0165 * (rdata->stick_y + sty_old) * player.dir.x;
		player.pos.z += 0.0165 * (rdata->stick_y + sty_old) * player.dir.z;
		player.pos.y += 0.03f * ud_old + (rdata->button & U_JPAD && TRUE);
		player.pos.y -= (rdata->button & D_JPAD && TRUE);

		sty_old = sty_old * 0.90f + 0.39f * rdata->stick_y;
		stx_old = 0.87f * stx_old + 0.012f * rdata->stick_x;
		ud_old = ud_old * 0.904f + (rdata->button & U_JPAD && TRUE) - (rdata->button & D_JPAD && TRUE);
	}
}

static void process_task( const Task* task ) {
	if( isRendering_Sys() ) {
		waitForGPU_Sys();
	}

	processTask_Engine( &engine, task );
	runGPU_Sys( &engine.dl );
}

static void handle_scene_action( const Time *time, void* arg ) {
	if( arg ) {}

	static Viewport		viewport;
	static Vector3f		randoms[BILLBOARDS_NUM];
	static Raw_Xformer	root;
	static Camera_Xformer	camera_xf;
	static Mesh_Xformer	worldmesh_xf, torusmesh_xf, torusmesh2_xf;
	static Lux_Xformer	lux_xf_r, lux_xf_b;
	static Billboard_Xformer	bb_xf[BILLBOARDS_NUM];
	static TreeNode		root_node, worldmesh_node, torusmesh_node, torusmesh2_node, lux_r_node, lux_b_node, camera_node;
	static TreeNode		bb_node[BILLBOARDS_NUM];

	static Matrix 		mv;
	static Vector3f		clearColor = {0.5f, 0.3f, 0.8f};
	static float		clearZ = 1.0f;
	static Task 		clearcfbTask = { (taskfptr)clearColorbufF_Engine, &clearColor };
	static Task 		clearzbufTask = { (taskfptr)clearZbufF_Engine, &clearZ };
	static Task 		sceneTask = { (taskfptr)sceneTaskF_Engine, &root_node };
	static bool 		initialized = false;


	if(!initialized) {
		root = createRaw_Engine( &engine );

		lux_xf_r = createLux_Engine( &engine );
		setColor_Lux( lux_xf_r.lux, 0.9f, 0.0f, 0.0f );

		lux_xf_b = createLux_Engine( &engine );
		setColor_Lux( lux_xf_b.lux, 0.0f, 0.0f, 0.9f );

		camera_xf = createCamera_Engine( &engine );
		setParams_Camera( camera_xf.camera, 63.74f, 1.33f, 1000.0f );

		worldmesh_xf = createMesh_Engine( &engine, getResource_ResMgr( "hyrulefield" ) );
		torusmesh_xf = createMesh_Engine( &engine, getResource_ResMgr( "torus" ) );
		torusmesh2_xf = createMesh_Engine( &engine, getResource_ResMgr( "torus" ) );

		// setup hierarchy
		init_TreeNode( &root_node, &root );
		init_TreeNode( &worldmesh_node, &worldmesh_xf );
		init_TreeNode( &torusmesh_node, &torusmesh_xf );
		init_TreeNode( &torusmesh2_node, &torusmesh2_xf );
		init_TreeNode( &lux_r_node, &lux_xf_r );
		init_TreeNode( &lux_b_node, &lux_xf_b );
		init_TreeNode( &camera_node, &camera_xf );

		attachChild_TreeNode( &root_node, &worldmesh_node );
		attachChild_TreeNode( &root_node, &torusmesh_node );
		attachChild_TreeNode( &torusmesh_node, &torusmesh2_node );
		attachChild_TreeNode( &torusmesh2_node, &lux_r_node );
		attachChild_TreeNode( &root_node, &lux_b_node );
		attachChild_TreeNode( &root_node, &camera_node );

		for(uint32_t i=0; i<BILLBOARDS_NUM; i++) {
			randoms[i].x = 0.004f * rand() * (rand() > 0x4000 ? 1 : -1);
			randoms[i].y = 0.001f * rand() * (rand() > 0x4000 ? 1 : -1);
			randoms[i].z = 0.004f * rand() * (rand() > 0x4000 ? 1 : -1);

			bb_xf[i] = createBillboard_Engine( &engine, getResource_ResMgr( "plane" ) );
			init_TreeNode( &bb_node[i], &bb_xf[i] );
			attachChild_TreeNode( &torusmesh2_node, &bb_node[i] );
		}


		// other stuff
		init_Viewport( &viewport, BUFFER_W, BUFFER_H );
		setViewport_Engine( &engine, &viewport );
		setZBuffer_Engine( &engine, getBuffer_Sys( BUFFER_Z ) );
		setCamera_Engine( &engine, &camera_xf );
		registerHandler_Sys( HANDLER_CONTROLLER, handle_controller );

		initialized = true;
	}

	/* camera */
	rotate_Matrix( &mv, player.angle, 0, 1, 0 );
	transformLocally_Xformer( camera_xf.xf, &mv );

	translate_Matrix( &mv, player.pos.x, player.pos.y, player.pos.z );
	transformLocally_Xformer( camera_xf.xf, &mv );

	/* lights */
	translate_Matrix( &mv, 0, 1, 0 );
	transformLocally_Xformer( lux_xf_r.xf, &mv );

	/* torus */
	rotate_Matrix( &mv, angle_torus, 1, 1, 1 );
	transformLocally_Xformer( torusmesh_xf.xf, &mv );

	translate_Matrix( &mv, 0, 160, 0 );
	transformLocally_Xformer( torusmesh_xf.xf, &mv );

	/* small torus */
	Vector3f pos0 = VECTOR3F_SET( 0.0f, -30.0f, 0.0f );
	Vector3f pos1 = VECTOR3F_SET( 0.0f, 30.0f, 0.0f );
	Vector3f pos = calculate_LERP3f( &pos0, &pos1, sinf( 3.2 * time->msec / 1000.f ) );
	//Vector3f pos = lerp3fCalculate( &pos0, &pos1, sinf( 3.2 * time.msec / 1000.f ) );
	scale_Matrix( &mv, 0.4f, 0.4f, 0.4f );
	transformLocally_Xformer( torusmesh2_xf.xf, &mv );
	translate_Matrix( &mv, pos.x, pos.y, pos.z );
	transformLocally_Xformer( torusmesh2_xf.xf, &mv );

	/* billboards */
	for(uint32_t i=0; i<BILLBOARDS_NUM; i++) {
		translate_Matrix( &mv, randoms[i].x, randoms[i].y, -randoms[i].z );
		transformLocally_Xformer( bb_xf[i].xf, &mv );
	}
#ifdef POSTPROC_BLUR
	setRenderTarget_Engine( &engine, getBuffer_Sys( BUFFER_AUX ) );
#else
	setRenderTarget_Engine( &engine, getBuffer_Sys( BUFFER_COLOR_DRAW ) );
#endif
	process_task( &clearcfbTask );
	process_task( &clearzbufTask );
	process_task( &sceneTask );
}
/*
static void HandleLogoAction(Time time, void* arg) {

	if(arg) {}
	if(time.sec) {}

	static Object2DTNode	logo;
	static Object2DTNode	sprite;
	static Task		imageTask = { (taskfptr)engineOverlayTaskF, &logo };
	static bool		initialized = false;

	if(!initialized) {
		object2dtnodeConstruct(&logo, resmgrGetResource("plane"));
		object2dtnodeConstruct(&sprite, resmgrGetResource("plane"));
		nodeAttach((Node*)&logo, (Node*)&sprite);
		initialized = true;
	}
	static Matrix mtx;

	matrixScale(&mtx, 1.0f/SCREEN_ASPECT_RATIO, 1.0f, 1.0f);
	tnodeTransform(&logo.tnode, &mtx);

	matrixTranslate(&mtx, -0.5f, -0.5f, 0.0f);
	tnodeTransform(&logo.tnode, &mtx); 

	matrixScale(&mtx, 0.5f, 0.5f, 1.0f);
	tnodeTransform(&sprite.tnode, &mtx);
	
	matrixTranslate(&mtx, 0.5f, 0.0f, 0.0f);
	tnodeTransform(&sprite.tnode, &mtx);

	ProcessTask(&imageTask);
#if 0
	static Texture		image; 
	static Object2D		obj2d;
	static TaskImageData	imageTaskData = {.obj2d = &obj2d};
	static Task		imageTask = { (taskfptr)engineImageTaskF, &imageTaskData };
	static bool		initialized = false;

	if(!initialized) {
		textureCopy(&image, resmgrGetResource("svlogo"));
		object2DConstruct(&obj2d, &image);
		initialized = true;
	}

	static Matrix mtx;
	static float sx = 1.0f;
	static float sy = 1.0f;
	matrixTranslate(&mtx, -0.5f, -0.5f*88/200, 0.0);
	tnodeTransform(&obj2d.tnode, &mtx); 
	if(time.sec<2)
		sx = sy = sinf((time.sec*1000 + time.msec)/(2000.0f*2/M_PI));
	matrixScale(&mtx, sx, sy, 1.0f);
	tnodeTransform(&obj2d.tnode, &mtx); 
	matrixTranslate(&mtx, 0.5f, 0.35f+0.5f*88/200, 0.0f);
	tnodeTransform(&obj2d.tnode, &mtx); 

	ProcessTask(&imageTask);
#endif
}
*/

/*
void rust_main(void);
uint32_t tezt(void);
float __floatundisf(unsigned long i) {
	return 0.0;
}
double __floatundidf(unsigned long i) {
	return 0.0;
}
int __mulodi4(int a, int b, int* overflow) {
}
*/

void mainproc( void* arg ) {
	if(arg) {}

	//rust_main();
	//while(tezt() < 0xFFFFF) {
	//}

	setAudioFrequency_Sys( 44100/2 );
	setAudioSource_Sys( (void*)((uintptr_t)getResource_ResMgr( "sound" ) + 44) );
	init_Engine(&engine);
	registerHandler_Sys( HANDLER_VBI, handle_vbi );

	/* actions */
	ActMgr actmgr;
	init_ActMgr(&actmgr);

	Action sceneAction;
	Time tss = {0,1,0};
	Time tsd = ACTMGR_TIME_INFINITY;	//{0,0,0};
	init_Action( &sceneAction, handle_scene_action, NULL, tss, tsd );
	addAction_ActMgr( &actmgr, &sceneAction );

	//Action logoAction;
	//Time tls = {0,100,0};
	//Time tld = {1000,200,0};
	//actionConstruct(&logoAction, HandleLogoAction, NULL, tls, tld);
	//actmgrAddAction(&actmgr, &logoAction);

	while(1) {
		Time current_time = getTime_Sys();
		process_ActMgr( &actmgr, &current_time );
#ifdef POSTPROC_BLUR
		const Surface* src = getBuffer_Sys( BUFFER_AUX );
		const Surface* dest = getBuffer_Sys( BUFFER_COLOR_DRAW );
		waitForGPU_Sys();
		postprocBlur(dest, src);
#else
		waitForGPU_Sys();
#endif
		swapBuffers_Sys();
	}
}

